from teslacam.config import Configuration
from teslacam.consts import NOTIFIERS

class NotificationService:
    def __init__(self, cfg: Configuration):
        if cfg.notifier:
            self.__notifier = NOTIFIERS[cfg.notifier](cfg)

    def notify(self, msg: str):
        if not self.__notifier:
            return

        try:
            self.__notifier.notify(msg)
        except:
            return


from abc import ABC, abstractmethod

from teslacam.config import Configuration
from teslacam.models import Clip

class Notifier(ABC):
    def __init__(self, cfg: Configuration):
        pass

    @abstractmethod
    def notify(self, message: str):
        pass

class Uploader(ABC):
    def __init__(self, cfg: Configuration):
        pass

    @abstractmethod
    def upload(self, clip: Clip) -> bool:
        """
        Uploads the clip.
        """
        pass


from azure.storage.blob import ContainerClient, BlobClient
from azure.core.exceptions import ResourceNotFoundError, ServiceRequestError

from teslacam.config import Configuration
from teslacam.contracts import Uploader
from teslacam.models import Clip

BLOB_STORAGE_CONFIG_KEY = "blobStorageUploader"

class BlobStorageUploader(Uploader):
    def __init__(self, cfg: Configuration):
        super().__init__(cfg)

        blob_cfg = cfg[BLOB_STORAGE_CONFIG_KEY]
        account_name = blob_cfg["accountName"]
        account_key = blob_cfg["accountKey"]
        container_name = blob_cfg["containerName"]

        self.__container_client = ContainerClient(f"https://{account_name}.blob.core.windows.net/",
            container_name, account_key, retry_total=1, connection_timeout=5)

    def upload(self, clip: Clip) -> bool:
        dir = f"{clip.date.year}/{clip.date.month}/{clip.date.day}" if clip.event != None else "recent"
        blob_name = f"{dir}/{clip.name}"

        blob = self.__container_client.get_blob_client(blob_name)

        try:
            blob.get_blob_properties()
            return False
        except ResourceNotFoundError:
            return self.__perform_upload(clip, blob)
        except ServiceRequestError:
            return False

    def __perform_upload(self, clip: Clip, blob: BlobClient) -> bool:
        try:
            with open(clip.path, "rb") as data:
                blob.upload_blob(data)
            return True
        except:
            return False


from os import path
from typing import Any, List, Optional

import yaml

from teslacam.enums import ClipType

CONFIG_PATH = "/etc/teslacam.yml"

class Configuration:
    def __init__(self, cfg: dict):
        self.__cfg = cfg

        self.__set_required_config()
        self.__set_optional_config()

    def __set_required_config(self):
        self.__tesla_cam_directory = self.__cfg["teslaCamDirectory"]
        self.__mount_directory = self.__cfg["mountDirectory"]
        self.__clip_types = [ClipType[type] for type in self.__cfg["clipTypes"]]

    def __set_optional_config(self):
        self.__uploader = self.__cfg.get("uploader")
        self.__notifier = self.__cfg.get("notifier")
        self.__upload_interval = self.__cfg.get("uploadInterval") or 30
        self.__last_event_clips_count = self.__cfg["lastEventClipsCount"] or 10

    def __getitem__(self, key: str) -> Any:
        return self.__cfg[key]

    @property
    def tesla_cam_directory(self) -> str:
        """
        Directory containing the TeslaCam directory.
        """
        return self.__tesla_cam_directory

    @property
    def mount_directory(self) -> bool:
        """
        Indicates whether the tesla_cam_directory should be mounted.
        """
        return self.__mount_directory

    @property
    def last_event_clips_count(self) -> int:
        """
        The amount of latest clips that should be uploaded for every event.
        """
        return self.__last_event_clips_count

    @property
    def uploader(self) -> Optional[str]:
        """
        The uploader to use.
        """
        return self.__uploader

    @property
    def clip_types(self) -> List[ClipType]:
        """
        Which clip types to upload.
        """
        return self.__clip_types

    @property
    def notifier(self) -> Optional[str]:
        """
        The notifier to use.
        """
        return self.__notifier

    @property
    def upload_interval(self) -> int:
        """
        The delay between upload job starts.
        """
        return self.__upload_interval

def load_config() -> Configuration:
    """
    Loads the application configuration from config.yml.
    """
    cfg_path = CONFIG_PATH if path.isfile(CONFIG_PATH) else "config.yml"

    with open(cfg_path) as file:
        cfg = yaml.load(file, Loader=yaml.FullLoader)
        return Configuration(cfg)


from chump import Application

from teslacam.config import Configuration
from teslacam.contracts import Notifier

PUSHOVER_CONFIG_KEY = "pushoverNotifier"

class PushoverNotifier(Notifier):
    def __init__(self, cfg: Configuration):
        super().__init__(cfg)

        pushover_cfg = cfg[PUSHOVER_CONFIG_KEY]
        apiToken = pushover_cfg["apiToken"]
        userKey = pushover_cfg["userKey"]

        application = Application(apiToken)
        self.__user = application.get_user(userKey)

    def notify(self, message: str):
        self.__user.send_message(message)


from datetime import datetime
from threading import Timer
from typing import List, Optional, Tuple

from teslacam.config import Configuration
from teslacam.consts import MIN_FILE_SIZE_BYTES, ONE_MEGABYTE, UPLOADERS
from teslacam.enums import ClipType
from teslacam.funcs import group_by
from teslacam.log import log
from teslacam.models import Clip
from teslacam.services.filesystem import FileSystem
from teslacam.services.notification import NotificationService

class UploadService:
    def __init__(self, cfg: Configuration, fs: FileSystem, notification: NotificationService):
        self.__cfg = cfg
        self.__fs = fs
        self.__notification = notification

        if cfg.uploader:
            self.__uploader = UPLOADERS[cfg.uploader](cfg)

        self.__timer: Optional[Timer] = None

    def start(self):
        """
        Starts a timer that will periodically upload TeslaCam clips.
        """
        if self.__timer is not None:
            return

        self.__timer = Timer(self.__cfg.upload_interval, self.__process_clips)
        self.__timer.start()

    def __process_clips(self):
        try:
            if (self.__cfg.mount_directory):
                self.__fs.mount_directory()

            total_uploaded = 0

            for type in self.__cfg.clip_types:
                total_uploaded += self.__process_of_type(type)

            if (self.__cfg.mount_directory):
                self.__fs.unmount_directory()
            
            log("Processing complete")

            if total_uploaded > 0:
                self.__notification.notify(f"Uploaded {total_uploaded} new TeslaCam clips")
        except:
            log("Processing failed, retrying later")

        self.__timer = None
        self.start()

    def __process_of_type(self, type: ClipType) -> int:
        log(f"Processing {str(type)} clips...")

        clips = self.__fs.read_clips(type)
        log(f"Found {len(clips)} clips")

        uploaded = 0

        (to_upload, to_delete) = self.__get_clips_to_upload(clips)

        for i, clip in enumerate(to_upload, start=1):
            log(f"Uploading clip '{clip.name}' ({format_size(clip)}) ({i}/{len(to_upload)})")
                
            if self.__uploader is not None and self.__uploader.upload(clip):
                clip.delete()
                uploaded += 1

        for clip in to_delete:
            log(f"Deleting clip '{clip.name}' ({format_size(clip)})")
            clip.delete()

        return uploaded

    def __get_clips_to_upload(self, clips: List[Clip]) -> Tuple[List[Clip], List[Clip]]:
        """
        Gets a tuple with the clips to upload and the clips to delete.
        """
        to_upload: List[Clip] = []
        to_skip: List[Clip] = []

        for event, event_clips in group_by(clips, lambda c: c.event).items():
            # Events less than 3 minutes ago will be skipped over for now
            if event is not None:
                diff = (datetime.today() - event).total_seconds() / 60

                if diff < 3.0:
                    to_skip.extend(event_clips)
                    break

            clips_by_date = group_by(event_clips, lambda c: c.date)
            dates = sorted(clips_by_date.keys())[-self.__cfg.last_event_clips_count:]

            clips_to_upload = [clip
                for date in dates
                for clip in clips_by_date[date]
                if clip.size >= MIN_FILE_SIZE_BYTES]

            to_upload.extend(clips_to_upload)

        to_delete = [clip
            for clip in clips
            if clip not in to_upload
            and clip not in to_skip]

        log(f"Will upload {len(to_upload)}, skip {len(to_skip)} and delete {len(to_delete)}")

        return (to_upload, to_delete)

def format_size(clip: Clip) -> str:
    return f"{round(clip.size / ONE_MEGABYTE, 2)} MB"


from os import path
from setuptools import setup, find_packages

here = path.dirname(__file__)

with open(path.join(here, "teslacam/VERSION"), encoding="utf-8") as file:
    version = file.read()

setup(
    name="teslacam_py",
    version=version,
    description="TeslaCam uploader",
    packages=find_packages(),
    python_requires=">=3.8",
    include_package_data=True,
    install_requires=[
        "pyyaml>=5.3",
        "azure-storage-blob>=12.1.0",
        "sh>=1.12.14",
        "flask>=1.1.1",
        "chump>=1.6.0"
    ],
    entry_points={"console_scripts": ["teslacam = teslacam.__main__:main"]}
)